import locale
import os
import platform
import sys

import site
import importlib


def get_execution_environment() -> dict:
    return dict(
        imports=get_loaded_modules(),
        os=get_os_info(),
        sys=get_sys_info(),
        sys_path=sys.path,
        site=get_site_info(),
        platform=get_platform_info(),
    )


def get_loaded_modules():
    return {
        dist.metadata["Name"]: dist.version
        for dist in importlib.metadata.distributions()
    }


def get_site_info():
    return {
        attr: getattr(site, attr)
        for attr in ["PREFIXES", "ENABLE_USER_SITE", "USER_SITE", "USER_BASE"]
    }


def get_platform_info():
    """
    Returns all available attributes from the platform module.
    """
    attributes = [
        attr
        for attr in dir(platform)
        if (not attr.startswith("_")) and callable(getattr(platform, attr))
    ]
    platform_info = {}
    for attr in attributes:
        try:
            platform_info[attr] = getattr(platform, attr)()
        except Exception as exc:
            # Not all attributes are available on all platforms.
            continue
    return platform_info


def get_os_info():
    values = {}
    for func in [
        "cwd",
        "egid",
        "euid",
        "gid",
        "groups",
        "login",
        "pgrp",
        "uid",
    ]:
        try:
            values[func] = getattr(os, "get" + func)()
        except (OSError, AttributeError) as exc:
            pass
    values["umask"] = oct(get_umask())
    values["name"] = os.name
    values["environ"] = {e: v for e, v in os.environ.items()}
    return values


def get_umask():
    # https://stackoverflow.com/questions/53227072/reading-umask-thread-safe
    current_value = os.umask(0)
    os.umask(current_value)
    return current_value


def get_sys_info():
    values = {}
    for attr in [
        "argv",
        "byteorder",
        "exec_prefix",
        "executable",
        "flags",
        "float_info",
        #      "maxint",
        "maxsize",
        "maxunicode",
        #   "meta_path",
    ]:
        values[attr] = getattr(sys, attr)
    for func in [
        "getdefaultencoding",
        "getfilesystemencoding",
        "getrecursionlimit",
    ]:
        try:
            values[func] = getattr(sys, func)()
        except (OSError, AttributeError) as exc:
            values[func] = exc
    return values


def localeconv():
    values = []
    for key, value in sorted(locale.localeconv().items()):
        if isinstance(value, bytes):
            value = value.decode("ascii", errors="replace")
        if key == "currency_symbol":
            value = repr(value)
        values.append("%s: %s" % (key, value))
    return values


def locale_module():
    values = []
    values.append("getdefaultlocale(): {}".format(locale.getdefaultlocale()))
    for category in [
        "LC_CTYPE",
        "LC_COLLATE",
        "LC_TIME",
        "LC_MONETARY",
        "LC_MESSAGES",
        "LC_NUMERIC",
    ]:
        values.append(
            "getlocale(locale.{}): {}".format(
                category, locale.getlocale(getattr(locale, category))
            )
        )
    return values



"""
This module defines the DataSet class with is used to manipulate datasets in DerivaML,
The intended use of this class is as a base class in DerivaML so all the methods documented here are
accessible via a DerivaML class instance.


"""

from bdbag.fetch.fetcher import fetch_single_file
from bdbag import bdbag_api as bdb
from collections import defaultdict
from deriva.core.ermrest_model import Table
from deriva.core.utils.core_utils import tag as deriva_tags, format_exception
from deriva.transfer.download.deriva_export import DerivaExport
from deriva.transfer.download.deriva_download import (
    DerivaDownloadConfigurationError,
    DerivaDownloadError,
    DerivaDownloadAuthenticationError,
    DerivaDownloadAuthorizationError,
    DerivaDownloadTimeoutError,
)

try:
    from icecream import ic
except ImportError:  # Graceful fallback if IceCream isn't installed.
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa

import json
import logging
from pathlib import Path
from pydantic import (
    validate_call,
    ConfigDict,
)
import requests

from tempfile import TemporaryDirectory, NamedTemporaryFile
from typing import Any, Callable, Optional, Iterable

from deriva_ml import DatasetBag
from .deriva_definitions import ML_SCHEMA, DerivaMLException, MLVocab, Status, RID
from .history import iso_to_snap
from .deriva_model import DerivaModel
from .database_model import DatabaseModel
from .dataset_aux_classes import (
    DatasetVersion,
    DatasetMinid,
    DatasetHistory,
    VersionPart,
    DatasetSpec,
)


class Dataset:
    """
    Class to manipulate a dataset.

    Attributes:
        dataset_table (Table): ERMRest table holding dataset information.
    """

    _Logger = logging.getLogger("deriva_ml")

    def __init__(self, model: DerivaModel, cache_dir: Path):
        self._model = model
        self._ml_schema = ML_SCHEMA
        self.dataset_table = self._model.schemas[self._ml_schema].tables["Dataset"]
        self._cache_dir = cache_dir
        self._logger = logging.getLogger("deriva_ml")

    def _is_dataset_rid(self, dataset_rid: RID, deleted: bool = False) -> bool:
        try:
            rid_info = self._model.catalog.resolve_rid(dataset_rid, self._model.model)
        except KeyError as _e:
            raise DerivaMLException(f"Invalid RID {dataset_rid}")

        # Got a dataset rid. Now check to see if its deleted or not.
        if deleted:
            return True
        else:
            return not list(rid_info.datapath.entities().fetch())[0]["Deleted"]

    def _insert_dataset_version(
        self,
        dataset_rid: RID,
        dataset_version: DatasetVersion,
        description: Optional[str] = "",
    ) -> RID:
        schema_path = self._model.catalog.getPathBuilder().schemas[self._ml_schema]
        version_path = schema_path.tables["Dataset_Version"]
        version_rid = version_path.insert(
            [
                {
                    "Dataset": dataset_rid,
                    "Version": str(dataset_version),
                    "Description": description,
                }
            ]
        )[0]["RID"]
        schema_path.tables["Dataset"].update(
            [{"RID": dataset_rid, "Version": version_rid}]
        )
        return version_rid

    def _bootstrap_versions(self):
        datasets = [ds["RID"] for ds in self.find_datasets()]
        ds_version = [
            {
                "Dataset": d,
                "Version": "0.1.0",
                "Description": "Dataset at the time of conversion to versioned datasets",
            }
            for d in datasets
        ]
        schema_path = self._model.catalog.getPathBuilder().schemas[self._ml_schema]
        version_path = schema_path.tables["Dataset_Version"]
        dataset_path = schema_path.tables["Dataset"]
        history = list(version_path.insert(ds_version))
        dataset_versions = [
            {"RID": h["Dataset"], "Version": h["Version"]} for h in history
        ]
        dataset_path.update(dataset_versions)

    def _synchronize_dataset_versions(self):
        datasets = [ds["RID"] for ds in self.find_datasets()]
        for ds in datasets:
            self.dataset_version(ds)
        schema_path = self._model.catalog.getPathBuilder().schemas[self._ml_schema]
        dataset_version_path = schema_path.tables["Dataset_Version"]
        # Get the maximum version number for each dataset.
        versions = {}
        for v in dataset_version_path.entities().fetch():
            if v["Version"] > versions.get("Dataset", DatasetVersion(0, 0, 0)):
                versions[v["Dataset"]] = v
        dataset_path = schema_path.tables["Dataset"]

        dataset_path.update(
            [
                {"RID": dataset, "Version": version["RID"]}
                for dataset, version in versions.items()
            ]
        )

    def dataset_history(self, dataset_rid: RID) -> list[DatasetHistory]:
        """Return a list of DatasetHistory objects representing the dataset

        Args:
            dataset_rid: A RID to the dataset for which history is to be fetched.

        Returns:
            A list of DatasetHistory objects which indicate the version-number, creation time, and bag instantiation of the dataset.
        """
        version_path = (
            self._model.catalog.getPathBuilder()
            .schemas[self._ml_schema]
            .tables["Dataset_Version"]
        )
        return [
            DatasetHistory(
                dataset_version=DatasetVersion.parse(v["Version"]),
                minid=v["Minid"],
                timestamp=v["RCT"],
                dataset_rid=dataset_rid,
                version_rid=v["RID"],
            )
            for v in version_path.filter(version_path.Dataset == dataset_rid)
            .entities()
            .fetch()
        ]

    @validate_call
    def dataset_version(self, dataset_rid: RID) -> DatasetVersion:
        """Retrieve the current version of the specified dataset_table.

        Given a rid, return the most recent version of the dataset. It is important to remember that this version
        captures the state of the catalog at the time the version was created, not the current state of the catalog.
        This means that its possible that the values associated with an object in the catalog may be different
        from the values of that object in the dataset.

        Args:
            dataset_rid: The RID of the dataset to retrieve the version for.

        Returns:
            A tuple with the semantic version of the dataset_table.
        """
        history = self.dataset_history(dataset_rid)
        if not history:
            return DatasetVersion(0, 1, 0)
        else:
            return max([h.dataset_version for h in self.dataset_history(dataset_rid)])

    def increment_dataset_version(
        self,
        dataset_rid: RID,
        component: VersionPart,
        description: Optional[str] = "",
        recurse: bool = False,
    ) -> DatasetVersion:
        """Increment the version of the specified dataset_table.

        Args:
          dataset_rid: RID to a dataset_table
          component: Which version of the dataset_table to increment.
          dataset_rid: RID of the dataset whose version is to be incremented.
          component: Major, Minor or Patch
          description: Description of the version update of the dataset_table.
          recurse: If True, increment the dataset_table recursively in a nested dataset.

        Returns:
          new semantic version of the dataset_table as a 3-tuple

        Raises:
          DerivaMLException: if provided RID is not to a dataset_table.
        """
        version = self.dataset_version(dataset_rid)
        new_version = version.increment_version(component)
        self._insert_dataset_version(dataset_rid, new_version, description=description)
        if recurse:
            for ds in self.list_dataset_children(dataset_rid, recurse=True):
                self.increment_dataset_version(ds, component, description=description)
        return new_version

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def create_dataset(
        self,
        type: str | list[str],
        description: str,
        execution_rid: Optional[RID] = None,
        version: Optional[DatasetVersion] = None,
    ) -> RID:
        """Create a new dataset_table from the specified list of RIDs.

        Args:
            type: One or more dataset_table types.  Must be a term from the DatasetType controlled vocabulary.
            description: Description of the dataset_table.
            execution_rid: Execution under which the dataset_table will be created.
            version: Version of the dataset_table.
            type: str | list[str]:
            description: str:


        Returns:
            New dataset_table RID.

        """

        version = version or DatasetVersion(0, 1, 0)

        type_path = (
            self._model.catalog.getPathBuilder()
            .schemas[self._ml_schema]
            .tables[MLVocab.dataset_type.value]
        )
        defined_types = list(type_path.entities().fetch())

        def check_dataset_type(dtype: str) -> bool:
            for term in defined_types:
                if dtype == term["Name"] or (
                    term["Synonyms"] and ds_type in term["Synonyms"]
                ):
                    return True
            return False

        # Create the entry for the new dataset_table and get its RID.
        ds_types = [type] if isinstance(type, str) else type
        pb = self._model.catalog.getPathBuilder()
        for ds_type in ds_types:
            if not check_dataset_type(ds_type):
                raise DerivaMLException(f"Dataset type must be a vocabulary term.")
        dataset_table_path = pb.schemas[self.dataset_table.schema.name].tables[
            self.dataset_table.name
        ]
        dataset_rid = dataset_table_path.insert(
            [
                {
                    "Description": description,
                    "Deleted": False,
                }
            ]
        )[0]["RID"]

        # Get the name of the association table between dataset_table and dataset_type.
        atable = next(
            self._model.schemas[self._ml_schema]
            .tables[MLVocab.dataset_type]
            .find_associations()
        ).name
        pb.schemas[self._ml_schema].tables[atable].insert(
            [
                {MLVocab.dataset_type: ds_type, "Dataset": dataset_rid}
                for ds_type in ds_types
            ]
        )
        if execution_rid is not None:
            pb.schemas[self._ml_schema].Dataset_Execution.insert(
                [{"Dataset": dataset_rid, "Execution": execution_rid}]
            )
        self._insert_dataset_version(dataset_rid, version)
        return dataset_rid

    @validate_call
    def delete_dataset(self, dataset_rid: RID, recurse: bool = False) -> None:
        """Delete a dataset_table from the catalog.

        Args:
            dataset_rid: RID of the dataset_table to delete.
            recurse: If True, delete the dataset_table along with any nested datasets. (Default value = False)
            dataset_rid: RID:
        """
        # Get association table entries for this dataset_table
        # Delete association table entries
        if not self._is_dataset_rid(dataset_rid):
            raise DerivaMLException("Dataset_rid is not a dataset.")

        if parents := self.list_dataset_parents(dataset_rid):
            raise DerivaMLException(
                f'Dataset_rid "{dataset_rid}" is in a nested dataset: {parents}.'
            )

        pb = self._model.catalog.getPathBuilder()
        dataset_path = pb.schemas[self.dataset_table.schema.name].tables[
            self.dataset_table.name
        ]

        rid_list = [dataset_rid] + (
            self.list_dataset_children(dataset_rid) if recurse else []
        )
        dataset_path.update([{"RID": r, "Deleted": True} for r in rid_list])

    def find_datasets(self, deleted: bool = False) -> Iterable[dict[str, Any]]:
        """Returns a list of currently available datasets.

        Arguments:
            deleted: If True, included the datasets that have been deleted.

        Returns:
             list of currently available datasets.
        """
        # Get datapath to all the tables we will need: Dataset, DatasetType and the association table.
        pb = self._model.catalog.getPathBuilder()
        dataset_path = pb.schemas[self.dataset_table.schema.name].tables[
            self.dataset_table.name
        ]
        atable = next(
            self._model.schemas[self._ml_schema]
            .tables[MLVocab.dataset_type]
            .find_associations()
        ).name
        ml_path = pb.schemas[self._ml_schema]
        atable_path = ml_path.tables[atable]

        if deleted:
            filtered_path = dataset_path
        else:
            filtered_path = dataset_path.filter(
                (dataset_path.Deleted == False) | (dataset_path.Deleted == None)
            )

        # Get a list of all the dataset_type values associated with this dataset_table.
        datasets = []
        for dataset in filtered_path.entities().fetch():
            ds_types = (
                atable_path.filter(atable_path.Dataset == dataset["RID"])
                .attributes(atable_path.Dataset_Type)
                .fetch()
            )
            datasets.append(
                dataset
                | {MLVocab.dataset_type: [ds[MLVocab.dataset_type] for ds in ds_types]}
            )
        return datasets

    def list_dataset_element_types(self) -> Iterable[Table]:
        """List the types of entities that can be added to a dataset_table.

        Returns:
          :return: An iterable of Table objects that can be included as an element of a dataset_table.
        """

        def domain_table(table: Table) -> bool:
            return (
                table.schema.name == self._model.domain_schema
                or table.name == self.dataset_table.name
            )

        return [
            t
            for a in self.dataset_table.find_associations()
            if domain_table(t := a.other_fkeys.pop().pk_table)
        ]

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def add_dataset_element_type(self, element: str | Table) -> Table:
        """A dataset_table is a heterogeneous collection of objects, each of which comes from a different table. This
        routine makes it possible to add objects from the specified table to a dataset_table.

        Args:
            element: Name or the table or table object that is to be added to the dataset_table.
            element: str | Table:

        Returns:
            The table object that was added to the dataset_table.
        """
        # Add table to map
        element_table = self._model.get_table(element)
        table = self._model.schemas[self._model.domain_schema].create_table(
            Table.define_association([self.dataset_table, element_table])
        )

        # self.model = self.catalog.getCatalogModel()
        self.dataset_table.annotations.update(self._generate_dataset_annotations())
        self._model.model.apply()
        return table

    @validate_call
    def list_dataset_members(
        self, dataset_rid: RID, recurse: bool = False
    ) -> dict[str, list[dict[str, Any]]]:
        """Return a list of entities associated with a specific dataset_table.

        Args:
            dataset_rid: param recurse: If this is a nested dataset_table, list the members of the contained datasets
            dataset_rid: RID:
            recurse:  (Default value = False)

        Returns:
            Dictionary of entities associated with a specific dataset_table.  Key is the table from which the elements
            were taken.
        """

        if not self._is_dataset_rid(dataset_rid):
            raise DerivaMLException(f"RID is not for a dataset_table: {dataset_rid}")

        # Look at each of the element types that might be in the dataset_table and get the list of rid for them from
        # the appropriate association table.
        members = defaultdict(list)
        pb = self._model.catalog.getPathBuilder()
        for assoc_table in self.dataset_table.find_associations():
            other_fkey = assoc_table.other_fkeys.pop()
            self_fkey = assoc_table.self_fkey
            target_table = other_fkey.pk_table
            member_table = assoc_table.table

            if (
                target_table.schema.name != self._model.domain_schema
                and target_table != self.dataset_table
            ):
                # Look at domain tables and nested datasets.
                continue
            if target_table == self.dataset_table:
                # find_assoc gives us the keys in the wrong position, so swap.
                self_fkey, other_fkey = other_fkey, self_fkey

            target_path = pb.schemas[target_table.schema.name].tables[target_table.name]
            member_path = pb.schemas[member_table.schema.name].tables[member_table.name]
            # Get the names of the columns that we are going to need for linking
            member_link = tuple(
                c.name for c in next(iter(other_fkey.column_map.items()))
            )
            path = pb.schemas[member_table.schema.name].tables[member_table.name].path
            path.filter(member_path.Dataset == dataset_rid)
            path.link(
                target_path,
                on=(
                    member_path.columns[member_link[0]]
                    == target_path.columns[member_link[1]]
                ),
            )
            target_entities = list(path.entities().fetch())
            members[target_table.name].extend(target_entities)
            if recurse and target_table == self.dataset_table:
                # Get the members for all the nested datasets and add to the member list.
                nested_datasets = [d["RID"] for d in target_entities]
                for ds in nested_datasets:
                    for k, v in self.list_dataset_members(ds, recurse=False).items():
                        members[k].extend(v)
        return dict(members)

    @validate_call
    def add_dataset_members(
        self,
        dataset_rid: RID,
        members: list[RID],
        validate: bool = True,
        description: Optional[str] = "",
    ) -> None:
        """Add additional elements to an existing dataset_table.

        Add new elements to an existing dataset. In addition to adding new members, the minor version number of the
        dataset is incremented and the description, if provide is applied to that new version.

        Args:
            dataset_rid: RID of dataset_table to extend or None if new dataset_table is to be created.
            members: List of RIDs of members to add to the  dataset_table.
            validate: Check rid_list to make sure elements are not already in the dataset_table.
            description: Markdown description of the updated dataset.
        """
        members = set(members)
        description = description or "Updated dataset via add_dataset_members"

        def check_dataset_cycle(member_rid, path=None):
            """

            Args:
              member_rid:
              path:  (Default value = None)

            Returns:

            """
            path = path or set(dataset_rid)
            return member_rid in path

        if validate:
            existing_rids = set(
                m["RID"]
                for ms in self.list_dataset_members(dataset_rid).values()
                for m in ms
            )
            if overlap := set(existing_rids).intersection(members):
                raise DerivaMLException(
                    f"Attempting to add existing member to dataset_table {dataset_rid}: {overlap}"
                )

        # Now go through every rid to be added to the data set and sort them based on what association table entries
        # need to be made.
        dataset_elements = {}
        association_map = {
            a.other_fkeys.pop().pk_table.name: a.table.name
            for a in self.dataset_table.find_associations()
        }
        # Get a list of all the types of objects that can be linked to a dataset_table.
        for m in members:
            try:
                rid_info = self._model.catalog.resolve_rid(m)
            except KeyError:
                raise DerivaMLException(f"Invalid RID: {m}")
            if rid_info.table.name not in association_map:
                raise DerivaMLException(
                    f"RID table: {rid_info.table.name} not part of dataset_table"
                )
            if rid_info.table == self.dataset_table and check_dataset_cycle(
                rid_info.rid
            ):
                raise DerivaMLException("Creating cycle of datasets is not allowed")
            dataset_elements.setdefault(rid_info.table.name, []).append(rid_info.rid)
        # Now make the entries into the association tables.
        pb = self._model.catalog.getPathBuilder()
        for table, elements in dataset_elements.items():
            schema_path = pb.schemas[
                self._ml_schema if table == "Dataset" else self._model.domain_schema
            ]
            fk_column = "Nested_Dataset" if table == "Dataset" else table
            if len(elements):
                # Find out the name of the column in the association table.
                schema_path.tables[association_map[table]].insert(
                    [{"Dataset": dataset_rid, fk_column: e} for e in elements]
                )
        self.increment_dataset_version(
            dataset_rid, VersionPart.minor, description=description
        )

    @validate_call
    def delete_dataset_members(
        self, dataset_rid: RID, members: list[RID], description=""
    ) -> None:
        """Remove elements to an existing dataset_table.

        Delete elements from an existing dataset. In addition to deleting members, the minor version number of the
        dataset is incremented and the description, if provide is applied to that new version.

        Args:
            dataset_rid: RID of dataset_table to extend or None if new dataset_table is to be created.
            members: List of RIDs of members to add to the  dataset_table.
            description: Markdown description of the updated dataset.
        """

        members = set(members)
        description = description or "Deletes dataset members"

        # Now go through every rid to be added to the data set and sort them based on what association table entries
        # need to be made.
        dataset_elements = {}
        association_map = {
            a.other_fkeys.pop().pk_table.name: a.table.name
            for a in self.dataset_table.find_associations()
        }
        # Get a list of all the types of objects that can be linked to a dataset_table.
        for m in members:
            try:
                rid_info = self._model.catalog.resolve_rid(m)
            except KeyError:
                raise DerivaMLException(f"Invalid RID: {m}")
            if rid_info.table.name not in association_map:
                raise DerivaMLException(
                    f"RID table: {rid_info.table.name} not part of dataset_table"
                )
            dataset_elements.setdefault(rid_info.table.name, []).append(rid_info.rid)
        # Now make the entries into the association tables.
        pb = self._model.catalog.getPathBuilder()
        for table, elements in dataset_elements.items():
            schema_path = pb.schemas[
                self._ml_schema if table == "Dataset" else self._model.domain_schema
            ]
            fk_column = "Nested_Dataset" if table == "Dataset" else table

            if len(elements):
                atable_path = schema_path.tables[association_map[table]]
                # Find out the name of the column in the association table.
                for e in elements:
                    entity = atable_path.filter(
                        (atable_path.Dataset == dataset_rid)
                        & (atable_path.columns[fk_column] == e),
                    )
                    entity.delete()
        self.increment_dataset_version(
            dataset_rid, VersionPart.minor, description=description
        )

    @validate_call
    def list_dataset_parents(self, dataset_rid: RID) -> list[RID]:
        """Given a dataset_table RID, return a list of RIDs of the parent datasets if this is included in a
        nested dataset.

        Args:
            dataset_rid: return: RID of the parent dataset_table.
            dataset_rid: RID:

        Returns:
            RID of the parent dataset_table.
        """
        if not self._is_dataset_rid(dataset_rid):
            raise DerivaMLException(
                f"RID: {dataset_rid} does not belong to dataset_table {self.dataset_table.name}"
            )
        # Get association table for nested datasets
        pb = self._model.catalog.getPathBuilder()
        atable_path = pb.schemas[self._ml_schema].Dataset_Dataset
        return [
            p["Dataset"]
            for p in atable_path.filter(atable_path.Nested_Dataset == dataset_rid)
            .entities()
            .fetch()
        ]

    @validate_call
    def list_dataset_children(self, dataset_rid: RID, recurse=False) -> list[RID]:
        """Given a dataset_table RID, return a list of RIDs of any nested datasets.

        Args:
            dataset_rid: A dataset_table RID.
            recurse: If True, return a list of RIDs of any nested datasets.

        Returns:
          list of RIDs of nested datasets.

        """
        children = [d["RID"] for d in self.list_dataset_members(dataset_rid)["Dataset"]]
        if recurse:
            for child in children.copy():
                children.extend(self.list_dataset_children(child, recurse=recurse))
        return children

    @staticmethod
    def _download_dataset_element(
        spath: str, dpath: str, table: Table
    ) -> list[dict[str, Any]]:
        """Return the download specification for the data object indicated by a path through the data model.

        Args:
          spath: Source path
          dpath: Destination path
          table: Table referenced to by the path

        Returns:
          The download specification that will retrieve that data from the catalog and place it into a BDBag.
        """
        exports = [
            {
                "processor": "csv",
                "processor_params": {
                    "query_path": f"/entity/{spath}?limit=none",
                    "output_path": dpath,
                },
            }
        ]

        # If this table is an asset table, then we need to output the files associated with the asset.
        asset_columns = {"Filename", "URL", "Length", "MD5", "Description"}
        if asset_columns.issubset({c.name for c in table.columns}):
            exports.append(
                {
                    "processor": "fetch",
                    "processor_params": {
                        "query_path": f"/attribute/{spath}/!(URL::null::)/url:=URL,length:=Length,filename:=Filename,md5:=MD5?limit=none",
                        "output_path": f"asset/{table.name}",
                    },
                }
            )
        return exports

    def _vocabulary_specification(
        self, writer: Callable[[str, str, Table], list[dict[str, Any]]]
    ) -> list[dict[str, Any]]:
        """

        Args:
          writer: Callable[[list[Table]]: list[dict[str: Any]]]:

        Returns:

        """
        vocabs = [
            table
            for s in self._model.schemas.values()
            for table in s.tables.values()
            if self._model.is_vocabulary(table)
        ]
        return [
            o
            for table in vocabs
            for o in writer(f"{table.schema.name}:{table.name}", table.name, table)
        ]

    def _domain_table_paths(
        self,
        graph: dict[Table, list[dict[Table, Any]]],
        spath: str = None,
        dpath: str = None,
        sprefix: str = "deriva-ml:Dataset/RID={Dataset_RID}",
        dprefix: str = "Dataset",
        nested: bool = False,
    ) -> list[tuple[str, str, Table]]:
        """Recursively walk over the domain schema graph and extend the current path.

        Args:
            graph: An undirected, acyclic graph of schema.  Represented as a dictionary whose name is the table name.
                and whose values are the child nodes of the table.
            spath: Source path so far
            dpath: Destination path so far
            sprefix: Initial path to be included.  Allows for nested datasets
            dprefix: Initial path to be included.  Allows for nested datasets
            nested: If true, skip initial data segment.

        Returns:
          A list of all the paths through the graph.  Each path is a list of tables.

        """
        source_path = spath or sprefix
        dest_path = dpath or dprefix
        paths = []
        for node, children in graph.items():
            if node.name == "Dataset":
                paths.append(
                    (
                        f"{sprefix}/(RID)=({self._ml_schema}:Dataset_Version:Dataset)",
                        f"{dprefix}/Dataset_Version",
                        self._model.schemas[self._ml_schema].tables["Dataset_Version"],
                    )
                )
                new_spath = sprefix
                new_dpath = dprefix

                if not nested:
                    paths.append((new_spath, new_dpath, node))
            else:
                new_spath = source_path + f"/{node.schema.name}:{node.name}"
                new_dpath = dest_path + f"/{node.name}"
                paths.append((new_spath, new_dpath, node))
            for child in children:
                paths.extend(
                    self._domain_table_paths(child, new_spath, new_dpath, nested=nested)
                )
        return paths

    def _table_paths(self, graph) -> list[tuple[str, str, Table]]:
        sprefix = "deriva-ml:Dataset/RID={Dataset_RID}"
        dprefix = "Dataset"
        dataset_dataset_table = self._model.schemas[self._ml_schema].tables[
            "Dataset_Dataset"
        ]
        table_paths = self._domain_table_paths(
            graph=graph, sprefix=sprefix, dprefix=dprefix
        )
        nested_sprefix = sprefix
        nested_dprefix = dprefix
        for i in range(self._dataset_nesting_depth()):
            nested_sprefix += f"/(RID)=(deriva-ml:Dataset_Dataset:Dataset)"
            nested_dprefix += f"/Dataset_Dataset"
            table_paths.append((nested_sprefix, nested_dprefix, dataset_dataset_table))
            nested_sprefix += f"/(Nested_Dataset)=(deriva-ml:Dataset:RID)"
            nested_dprefix += f"/Dataset"
            table_paths.append((nested_sprefix, nested_dprefix, self.dataset_table))
            # Get CSV for nested datasets.
            table_paths.extend(
                self._domain_table_paths(
                    graph, sprefix=nested_sprefix, dprefix=nested_dprefix, nested=True
                )
            )
        return table_paths

    def _dataset_nesting_depth(self):
        """Determine the maximum dataset nesting depth in the current catalog.

        Returns:

        """

        def children_depth(
            dataset_rid: RID, nested_datasets: dict[RID, list[RID]]
        ) -> int:
            try:
                children = nested_datasets[dataset_rid]
                return (
                    max(map(lambda x: children_depth(x, nested_datasets), children)) + 1
                    if children
                    else 1
                )
            except KeyError:
                return 0

        # Build up the dataset_table nesting graph...
        pb = (
            self._model.catalog.getPathBuilder()
            .schemas[self._ml_schema]
            .tables["Dataset_Dataset"]
        )
        nested_dataset = defaultdict(list)
        for ds in pb.entities().fetch():
            nested_dataset[ds["Dataset"]].append(ds["Nested_Dataset"])
        return (
            max(map(lambda d: children_depth(d, dict(nested_dataset)), nested_dataset))
            if nested_dataset
            else 0
        )

    def _schema_graph(
        self, node: Table, visited_nodes: Optional[set] = None
    ) -> dict[Table, list[dict[Table, list]]]:
        """Generate an undirected, acyclic graph of domain schema. We do this by traversing the schema foreign key
        relationships.  We stop when we hit the deriva-ml schema or when we reach a node that we have already seen.

        Nested datasets need to be unfolded

        Args:
          node: Current (starting) node in the graph.
          visited_nodes: param nested_dataset: Are we in a nested dataset_table, (i.e. have we seen the DataSet table)?

        Returns:
            Graph of the schema, starting from node.
        """

        visited_nodes = visited_nodes or set()
        graph = {node: []}

        def include_node(child: Table) -> bool:
            """Indicate if the table should be included in the graph.

            Include node in the graph if it's not a loopback from fk<-> referred_by, you have not already been to the
            node.
            """
            return (
                child != node
                and child not in visited_nodes
                and child.schema.name == self._model.domain_schema
            )

        # Get all the tables reachable from the end of the path avoiding loops from T1<->T2 via referenced_by
        nodes = {fk.pk_table for fk in node.foreign_keys if include_node(fk.pk_table)}
        nodes |= {fk.table for fk in node.referenced_by if include_node(fk.table)}
        for t in nodes:
            new_visited_nodes = visited_nodes.copy()
            new_visited_nodes.add(t)
            if self._model.is_vocabulary(t):
                # If the end of the path is a vocabulary table, we are at a terminal node in the ERD, so stop
                continue
            # Get all the paths that extend the current path
            graph[node].append(self._schema_graph(t, new_visited_nodes))
        return graph

    def _dataset_specification(
        self, writer: Callable[[str, str, Table], list[dict[str, Any]]]
    ) -> list[dict[str, Any]]:
        """Output a download/export specification for a dataset_table.  Each element of the dataset_table will be placed in its own dir
        The top level data directory of the resulting BDBag will have one subdirectory for element type. the subdirectory
        will contain the CSV indicating which elements of that type are present in the dataset_table, and then there will be a
        subdirectories for each object that is reachable from the dataset_table members.

        To simplify reconstructing the relationship between tables, the CVS for each
        The top level data directory will also contain a subdirectory for any controlled vocabularies used in the dataset_table.
        All assets will be placed into a directory named asset in a subdirectory with the asset table name.

        For example, consider a dataset_table that consists of two element types, T1 and T2. T1 has foreign key relationships to
        objects in tables T3 and T4.  There are also two controlled vocabularies, CV1 and CV2.  T2 is an asset table
        which has two asset in it. The layout of the resulting bdbag would be:
              data
                CV1/
                    cv1.csv
                CV2/
                    cv2.csv
                Dataset/
                    T1/
                        t1.csv
                        T3/
                            t3.csv
                        T4/
                            t4.csv
                    T2/
                        t2.csv
                asset/
                  T2
                    f1
                    f2

        Args:
          writer: Callable[[list[Table]]: list[dict[str:  Any]]]:

        Returns:
            A dataset_table specification.
        """
        element_spec = []
        for path in self._table_paths(self._schema_graph(self.dataset_table)):
            element_spec.extend(writer(*path))
        return self._vocabulary_specification(writer) + element_spec

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def download_dataset_bag(
        self,
        dataset: DatasetSpec,
        execution_rid: Optional[RID] = None,
    ) -> DatasetBag:
        """Download a dataset onto the local file system.  Create a MINID for the dataset if one doesn't already exist.

        Args:
            dataset: Specification of the dataset to be downloaded.
            execution_rid: Execution RID for the dataset.

        Returns:
            Tuple consisting of the path to the dataset, the RID of the dataset that was downloaded and the MINID
            for the dataset.
        """
        if (
            execution_rid
            and self._model.catalog.resolve_rid(execution_rid).table.name != "Execution"
        ):
            raise DerivaMLException(f"RID {execution_rid} is not an execution")
        minid = self.get_dataset_minid(dataset.rid, dataset.version)

        bag_path = (
            self._materialize_dataset_bag(minid, execution_rid=execution_rid)
            if dataset.materialize
            else self._download_dataset_bag(minid)
        )
        return DatabaseModel.register(minid, bag_path).get_dataset()

    def _version_snapshot(self, dataset_rid: RID, version: DatasetVersion) -> str:
        version_timestamp = [
            h.timestamp
            for h in self.dataset_history(dataset_rid)
            if h.dataset_version == version
        ][0]
        snaptime = iso_to_snap(version_timestamp.isoformat())
        return f"{self._model.catalog.catalog_id}@{snaptime}"

    def _create_dataset_minid(
        self, dataset_rid: RID, dataset_version: DatasetHistory
    ) -> str:
        with TemporaryDirectory() as tmp_dir:
            # Generate a download specification file for the current catalog schema. By default, this spec
            # will generate a minid and place the bag into S3 storage.
            spec_file = f"{tmp_dir}/download_spec.json"
            with open(spec_file, "w", encoding="utf-8") as ds:
                json.dump(self._generate_dataset_download_spec(), ds)
            try:
                # Generate the bag and put into S3 storage.
                exporter = DerivaExport(
                    host=self._model.catalog.deriva_server.server,
                    catalog_id=self._version_snapshot(
                        dataset_rid, dataset_version.dataset_version
                    ),
                    config_file=spec_file,
                    output_dir=tmp_dir,
                    defer_download=True,
                    timeout=(10, 300),
                    envars={"Dataset_RID": dataset_rid},
                )
                minid_page_url = exporter.export()[0]  # Get the MINID launch page
            except (
                DerivaDownloadError,
                DerivaDownloadConfigurationError,
                DerivaDownloadAuthenticationError,
                DerivaDownloadAuthorizationError,
                DerivaDownloadTimeoutError,
            ) as e:
                raise DerivaMLException(format_exception(e))
            # Update version table with MINID.
            version_path = (
                self._model.catalog.getPathBuilder()
                .schemas[self._ml_schema]
                .tables["Dataset_Version"]
            )
            version_path.update(
                [{"RID": dataset_version.version_rid, "Minid": minid_page_url}]
            )
        return minid_page_url

    def get_dataset_minid(
        self,
        dataset_rid: RID,
        version: Optional[DatasetVersion] = None,
        create: bool = True,
    ) -> DatasetMinid:
        """Return a MINID to the specified dataset.  If no version is specified, use the latest.

        Args:
            dataset_rid: RID of the dataset.
            version: Version of the dataset.
            create: Create a new MINID if one doesn't already exist.

        Returns:
            New or existing MINID for the dataset.
        """
        if dataset_rid.startswith("minid"):
            minid_url = f"https://identifiers.org/{dataset_rid}"
        elif dataset_rid.startswith("http"):
            minid_url = dataset_rid
        else:
            if not any([dataset_rid == ds["RID"] for ds in self.find_datasets()]):
                raise DerivaMLException(f"RID {dataset_rid} is not a dataset_table")

            # Get the history record for the version we are looking for.
            version = version or self.dataset_version(dataset_rid)
            dataset_version = [
                v
                for v in self.dataset_history(dataset_rid)
                if v.dataset_version == str(version)
            ][0]
            if not dataset_version:
                raise DerivaMLException(
                    f"Version {str(version)} does not exist for RID {dataset_rid}"
                )
            minid_url = dataset_version.minid
            if not minid_url:
                if not create:
                    raise DerivaMLException(
                        f"Minid for dataset {dataset_rid} doesn't exist"
                    )
                self._logger.info("Creating new MINID for dataset %s", dataset_rid)
                minid_url = self._create_dataset_minid(dataset_rid, dataset_version)
            # If provided a MINID, use the MINID metadata to get the checksum and download the bag.
        r = requests.get(minid_url, headers={"accept": "application/json"})
        return DatasetMinid(dataset_version=version, **r.json())

    def _download_dataset_bag(self, minid: DatasetMinid) -> Path:
        """Given a RID to a dataset_table, or a MINID to an existing bag, download the bag file, extract it and validate
        that all the metadata is correct

        Args:
            minid: The RID of a dataset_table or a minid to an existing bag.
        Returns:
            the location of the unpacked and validated dataset_table bag and the RID of the bag and the bag MINID
        """

        # Check to see if we have an existing idempotent materialization of the desired bag. If so, then just reuse
        # it.  If not, then we need to extract the contents of the archive into our cache directory.
        bag_dir = self._cache_dir / f"{minid.dataset_rid}_{minid.checksum}"
        if bag_dir.exists():
            bag_path = (bag_dir / f"Dataset_{minid.dataset_rid}").as_posix()
        else:
            bag_dir.mkdir(parents=True, exist_ok=True)
            with NamedTemporaryFile(
                delete=False, suffix=f"Dataset_{minid.dataset_rid}.zip"
            ) as zip_file:
                archive_path = fetch_single_file(minid.bag_url, zip_file.name)
                bag_path = bdb.extract_bag(archive_path, bag_dir.as_posix())
            bdb.validate_bag_structure(bag_path)
        return Path(bag_path)

    def _materialize_dataset_bag(
        self,
        minid: DatasetMinid,
        execution_rid: Optional[RID] = None,
    ) -> Path:
        """Materialize a dataset_table bag into a local directory

        Args:
            minid: A MINID to an existing bag or a RID of the dataset_table that should be downloaded.

        Returns:
            A tuple containing the path to the bag, the RID of the bag, and the MINID to the bag.
        """

        def update_status(status: Status, msg: str) -> None:
            self._model.catalog.getPathBuilder().schemas[
                self._ml_schema
            ].Execution.update(
                [
                    {
                        "RID": execution_rid,
                        "Status": status.value,
                        "Status_Detail": msg,
                    }
                ]
            )
            self._logger.info(msg)

        def fetch_progress_callback(current, total):
            msg = f"Materializing bag: {current} of {total} file(s) downloaded."
            if execution_rid:
                update_status(Status.running, msg)
            return True

        def validation_progress_callback(current, total):
            msg = f"Validating bag: {current} of {total} file(s) validated."
            if execution_rid:
                update_status(Status.running, msg)
            return True

        # request metadata
        bag_path = self._download_dataset_bag(minid)
        bag_dir = bag_path.parent
        validated_check = bag_dir / "validated_check.txt"

        # If this bag has already been validated, our work is done.  Otherwise, materialize the bag.
        if not validated_check.exists():
            bdb.materialize(
                bag_path.as_posix(),
                fetch_callback=fetch_progress_callback,
                validation_callback=validation_progress_callback,
            )
            validated_check.touch()
        return Path(bag_path)

    def _export_outputs(self) -> list[dict[str, Any]]:
        """Return and output specification for the datasets in the provided model

        Returns:
          An export specification suitable for Chaise.
        """

        def writer(spath: str, dpath: str, table: Table) -> list[dict[str, Any]]:
            """

            Args:
              spath: list[Table]:
              dpath: list[Table]:
              table: Table

            Returns:
                An export specification suitable for Chaise.
            """
            return self._export_dataset_element(spath, dpath, table)

        # Export specification is a specification for the datasets, plus any controlled vocabulary
        return [
            {
                "source": {"api": False, "skip_root_path": True},
                "destination": {"type": "env", "params": {"query_keys": ["snaptime"]}},
            },
            {
                "source": {"api": "entity"},
                "destination": {
                    "type": "env",
                    "params": {"query_keys": ["RID", "Description"]},
                },
            },
            {
                "source": {"api": "schema", "skip_root_path": True},
                "destination": {"type": "json", "name": "schema"},
            },
        ] + self._dataset_specification(writer)

    def _processor_params(self) -> list[dict[str, Any]]:
        """
        Returns:
          a download specification for the datasets in the provided model.

        """

        def writer(spath: str, dpath: str, table: Table) -> list[dict[str, Any]]:
            """

            Args:
              spath:
              dpath:
              table: Table

            Returns:

            """
            return self._download_dataset_element(spath, dpath, table)

        # Download spec is the spec for any controlled vocabulary and for the dataset_table.
        return [
            {
                "processor": "json",
                "processor_params": {"query_path": f"/schema", "output_path": "schema"},
            }
        ] + self._dataset_specification(writer)

    @staticmethod
    def _export_dataset_element(
        spath: str, dpath: str, table: Table
    ) -> list[dict[str, Any]]:
        """Given a path in the data model, output an export specification for the path taken to get to the current table.

        Args:
          spath: Source path
          dpath: Destination path
          table: Table referenced to by the path

        Returns:
          The export specification that will retrieve that data from the catalog and place it into a BDBag.
        """
        # The table is the last element of the path.  Generate the ERMRest query by converting the list of tables
        # into a path in the form of /S:T1/S:T2/S:Table
        # Generate the destination path in the file system using just the table names.

        exports = [
            {
                "source": {"api": "entity", "path": spath},
                "destination": {"name": dpath, "type": "csv"},
            }
        ]

        # If this table is an asset table, then we need to output the files associated with the asset.
        asset_columns = {"Filename", "URL", "Length", "MD5", "Description"}
        if asset_columns.issubset({c.name for c in table.columns}):
            exports.append(
                {
                    "source": {
                        "api": "attribute",
                        "path": f"{spath}/!(URL::null::)/url:=URL,length:=Length,filename:=Filename,md5:=MD5",
                    },
                    "destination": {"name": f"asset/{table.name}", "type": "fetch"},
                }
            )
        return exports

    def _generate_dataset_download_spec(self) -> dict[str, Any]:
        """

        Returns:
        """
        s3_target = "s3://eye-ai-shared"
        minid_test = False

        return {
            "env": {"Dataset_RID": "{Dataset_RID}"},
            "bag": {
                "bag_name": "Dataset_{Dataset_RID}",
                "bag_algorithms": ["md5"],
                "bag_archiver": "zip",
                "bag_metadata": {},
                "bag_idempotent": True,
            },
            "post_processors": [
                {
                    "processor": "cloud_upload",
                    "processor_params": {
                        "acl": "public-read",
                        "target_url": s3_target,
                    },
                },
                {
                    "processor": "identifier",
                    "processor_params": {
                        "test": minid_test,
                        "env_column_map": {
                            "Dataset_RID": "{RID}@{snaptime}",
                            "Description": "{Description}",
                        },
                    },
                },
            ],
            "catalog": {
                "host": f"{self._model.catalog.deriva_server.scheme}://{self._model.catalog.deriva_server.server}",
                "catalog_id": f"{self._model.catalog.catalog_id}",
                "query_processors": [
                    {
                        "processor": "env",
                        "processor_params": {
                            "output_path": "Dataset",
                            "query_keys": ["snaptime"],
                            "query_path": "/",
                        },
                    },
                    {
                        "processor": "env",
                        "processor_params": {
                            "query_path": "/entity/M:=deriva-ml:Dataset/RID={Dataset_RID}?limit=none",
                            "output_path": "Dataset",
                            "query_keys": ["RID", "Description"],
                        },
                    },
                ]
                + self._processor_params(),
            },
        }

    def dataset_visible_columns(self) -> dict[str, Any]:
        dataset_table = self._model.schemas["deriva-ml"].tables["Dataset"]
        rcb_name = next(
            [fk.name[0].name, fk.name[1]]
            for fk in dataset_table.foreign_keys
            if fk.name[1] == "Dataset_RCB_fkey"
        )
        rmb_name = next(
            [fk.name[0].name, fk.name[1]]
            for fk in dataset_table.foreign_keys
            if fk.name[1] == "Dataset_RMB_fkey"
        )
        return {
            "*": [
                "RID",
                "Description",
                {
                    "display": {
                        "markdown_pattern": "[Annotate Dataset](https://www.eye-ai.org/apps/grading-interface/main?dataset_rid={{{RID}}}){: .btn}"
                    },
                    "markdown_name": "Annotation App",
                },
                rcb_name,
                rmb_name,
            ],
            "detailed": [
                "RID",
                "Description",
                {
                    "source": [
                        {"inbound": ["deriva-ml", "Dataset_Dataset_Type_Dataset_fkey"]},
                        {
                            "outbound": [
                                "deriva-ml",
                                "Dataset_Dataset_Type_Dataset_Type_fkey",
                            ]
                        },
                        "RID",
                    ],
                    "markdown_name": "Dataset Types",
                },
                {
                    "display": {
                        "markdown_pattern": "[Annotate Dataset](https://www.eye-ai.org/apps/grading-interface/main?dataset_rid={{{RID}}}){: .btn}"
                    },
                    "markdown_name": "Annotation App",
                },
                rcb_name,
                rmb_name,
            ],
            "filter": {
                "and": [
                    {"source": "RID"},
                    {"source": "Description"},
                    {
                        "source": [
                            {
                                "inbound": [
                                    "deriva-ml",
                                    "Dataset_Dataset_Type_Dataset_fkey",
                                ]
                            },
                            {
                                "outbound": [
                                    "deriva-ml",
                                    "Dataset_Dataset_Type_Dataset_Type_fkey",
                                ]
                            },
                            "RID",
                        ],
                        "markdown_name": "Dataset Types",
                    },
                    {
                        "source": [{"outbound": rcb_name}, "RID"],
                        "markdown_name": "Created By",
                    },
                    {
                        "source": [{"outbound": rmb_name}, "RID"],
                        "markdown_name": "Modified By",
                    },
                ]
            },
        }

    def _dataset_visible_fkeys(self) -> dict[str, Any]:
        def fkey_name(fk):
            return [fk.name[0].name, fk.name[1]]

        dataset_table = self._model.schemas["deriva-ml"].tables["Dataset"]

        source_list = [
            {
                "source": [
                    {"inbound": fkey_name(fkey.self_fkey)},
                    {"outbound": fkey_name(other_fkey := fkey.other_fkeys.pop())},
                    "RID",
                ],
                "markdown_name": other_fkey.pk_table.name,
            }
            for fkey in dataset_table.find_associations(max_arity=3, pure=False)
        ]
        return {"detailed": source_list}

    def _generate_dataset_annotations(self) -> dict[str, Any]:
        return {
            deriva_tags.export_fragment_definitions: {
                "dataset_export_outputs": self._export_outputs()
            },
            deriva_tags.visible_columns: self.dataset_visible_columns(),
            deriva_tags.visible_foreign_keys: self._dataset_visible_fkeys(),
            deriva_tags.export_2019: {
                "detailed": {
                    "templates": [
                        {
                            "type": "BAG",
                            "outputs": [{"fragment_key": "dataset_export_outputs"}],
                            "displayname": "BDBag Download",
                            "bag_idempotent": True,
                            "postprocessors": [
                                {
                                    "processor": "identifier",
                                    "processor_params": {
                                        "test": False,
                                        "env_column_map": {
                                            "Dataset_RID": "{RID}@{snaptime}",
                                            "Description": "{Description}",
                                        },
                                    },
                                }
                            ],
                        },
                        {
                            "type": "BAG",
                            "outputs": [{"fragment_key": "dataset_export_outputs"}],
                            "displayname": "BDBag to Cloud",
                            "bag_idempotent": True,
                            "postprocessors": [
                                {
                                    "processor": "cloud_upload",
                                    "processor_params": {
                                        "acl": "public-read",
                                        "target_url": "s3://eye-ai-shared/",
                                    },
                                },
                                {
                                    "processor": "identifier",
                                    "processor_params": {
                                        "test": False,
                                        "env_column_map": {
                                            "Dataset_RID": "{RID}@{snaptime}",
                                            "Description": "{Description}",
                                        },
                                    },
                                },
                            ],
                        },
                    ]
                }
            },
        }



from derivaml_test import TestDerivaML
from deriva_ml import DatasetSpec
from pathlib import Path


class TestDownload(TestDerivaML):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def test_download(self):
        double_nested_dataset, nested_datasets, datasets = self.create_nested_dataset()
        bag = self.ml_instance.download_dataset_bag(
            DatasetSpec(
                rid=double_nested_dataset,
                version=self.ml_instance.dataset_version(double_nested_dataset),
            )
        )

        # The datasets in the bag should be all the datasets we started with.
        self.assertEqual(
            set([double_nested_dataset] + nested_datasets + datasets),
            {k for k in bag.model.bag_rids.keys()},
        )

        # Children of top level bag should be in datasets variable
        self.assertEqual(
            set(nested_datasets), {ds.dataset_rid for ds in bag.list_dataset_children()}
        )

        self.assertEqual(
            set(nested_datasets + datasets),
            {ds.dataset_rid for ds in bag.list_dataset_children(recurse=True)},
        )

        # Check to see if all of the files have been downloaded.
        files = [Path(r["Filename"]) for r in bag.get_table_as_dict("Image")]
        for f in files:
            self.assertTrue(f.exists())



"""
The module implements the sqllite interface to a set of directories representing a dataset bag.
"""

from collections import defaultdict
from copy import copy
from typing import Any, Generator, TYPE_CHECKING, Optional

import pandas as pd
from pydantic import validate_call
from .deriva_definitions import RID

if TYPE_CHECKING:
    from .database_model import DatabaseModel


class DatasetBag:
    """DatasetBag is a class that manages a materialized bag.  It is created from a locally materialized BDBag for a
    dataset_table, which is created either by DerivaML.create_execution, or directly by calling DerivaML.download_dataset.

    A general a bag may contain multiple datasets, if the dataset is nested. The DatasetBag is used to represent only
    one of the datasets in the bag.

    All the metadata associated with the dataset is stored in a SQLLite database that can be queried using SQL.

    Attributes
        dataset_rid (RID): RID for the specified dataset
        version: The version of the dataset
        model (DatabaseModel): The Database model that has all the catalog metadata associated with this dataset.
            database:
        dbase (Connection): connection to the sqlite database holding table values
        domain_schema (str): Name of the domain schema
    """

    # Validate call
    def __init__(
        self, database_model: "DatabaseModel", dataset_rid: Optional[RID]
    ) -> None:
        """
        Initialize a DatasetBag instance.

        Args:
            database_model: Database version of the bag.
        """

        self.model = database_model
        self.database = self.model.dbase

        self.dataset_rid = dataset_rid or self.model.dataset_rid
        self.model.rid_lookup(
            dataset_rid
        )  # Check to make sure that this dataset is in the bag.

        self.version = self.model.dataset_version(self.dataset_rid)
        self._dataset_table = self.model.dataset_table

    def __repr__(self) -> str:
        return f"<deriva_ml.DatasetBag object {self.dataset_rid} at {hex(id(self))}>"

    def list_tables(self) -> list[str]:
        """List the names of the tables in the catalog

        Returns:
            A list of table names.  These names are all qualified with the Deriva schema name.
        """
        return self.model.list_tables()

    def get_table(self, table: str) -> Generator[tuple, None, None]:
        """Get the contents of the specified table as a set of tuples"""
        return self.model.get_table(table)

    def get_table_as_dataframe(self, table: str) -> pd.DataFrame:
        """Get the contents of the specified table as a dataframe"""
        return self.model.get_table_as_dataframe(table)

    def get_table_as_dict(self, table: str) -> Generator[dict[str, Any], None, None]:
        """Get the contents of the specified table as a dictionary"""
        return self.model.get_table_as_dict(table)

    @validate_call
    def list_dataset_members(self, recurse: bool = False) -> dict[str, list[tuple]]:
        """Return a list of entities associated with a specific _dataset_table.

        Args:
           recurse:  (Default value = False)

        Returns:
            Dictionary of entities associated with a specific _dataset_table.  Key is the table from which the elements
            were taken.
        """

        # Look at each of the element types that might be in the _dataset_table and get the list of rid for them from
        # the appropriate association table.
        members = defaultdict(list)
        for assoc_table in self._dataset_table.find_associations():
            other_fkey = assoc_table.other_fkeys.pop()
            self_fkey = assoc_table.self_fkey
            target_table = other_fkey.pk_table
            member_table = assoc_table.table

            if (
                target_table.schema.name != self.model.domain_schema
                and target_table != self._dataset_table
            ):
                # Look at domain tables and nested datasets.
                continue
            if target_table == self._dataset_table:
                # find_assoc gives us the keys in the wrong position, so swap.
                self_fkey, other_fkey = other_fkey, self_fkey
            sql_target = self.model.normalize_table_name(target_table.name)
            sql_member = self.model.normalize_table_name(member_table.name)

            # Get the names of the columns that we are going to need for linking
            member_link = tuple(
                c.name for c in next(iter(other_fkey.column_map.items()))
            )

            with self.database as db:
                sql_cmd = (
                    f'SELECT * FROM "{sql_member}" '
                    f'JOIN "{sql_target}" ON "{sql_member}".{member_link[0]} = "{sql_target}".{member_link[1]} '
                    f'WHERE "{self.dataset_rid}" = "{sql_member}".Dataset;'
                )
                target_entities = db.execute(sql_cmd).fetchall()
                members[target_table.name].extend(target_entities)

            target_entities = []  # path.entities().fetch()
            members[target_table.name].extend(target_entities)
            if recurse and target_table.name == self._dataset_table:
                # Get the members for all the nested datasets and add to the member list.
                nested_datasets = [d["RID"] for d in target_entities]
                for ds in nested_datasets:
                    for k, v in DatasetBag.list_dataset_members(
                        ds, recurse=False
                    ).items():
                        members[k].extend(v)
        return dict(members)

    # @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def list_dataset_children(self, recurse: bool = False) -> list["DatasetBag"]:
        """Given a _dataset_table RID, return a list of RIDs of any nested datasets.

        Returns:
          list of RIDs of nested datasets.

        """
        ds_table = self.model.normalize_table_name("Dataset")
        nds_table = self.model.normalize_table_name("Dataset_Dataset")
        dv_table = self.model.normalize_table_name("Dataset_Version")
        with self.database as db:
            sql_cmd = (
                f'SELECT  "{nds_table}".Nested_Dataset, "{dv_table}".Version '
                f'FROM "{nds_table}" JOIN "{dv_table}" JOIN "{ds_table}" on '
                f'"{ds_table}".Version == "{dv_table}".RID AND '
                f'"{nds_table}".Nested_Dataset == "{ds_table}".RID '
                f'where "{nds_table}".Dataset == "{self.dataset_rid}"'
            )
            nested = [
                DatasetBag(self.model, r[0]) for r in db.execute(sql_cmd).fetchall()
            ]

        result = copy(nested)
        if recurse:
            for child in nested:
                result.extend(child.list_dataset_children(recurse))
        return result


# Add annotations after definition to deal with forward reference issues in pydantic

DatasetBag.list_dataset_children = validate_call(
    config=dict(arbitrary_types_allowed=True),
    validate_return=True,
)(DatasetBag.list_dataset_children)



"""
`deriva_ml_base.py` is the core module for the Deriva ML project.  This module implements the DerivaML class, which is
the primary interface to the Deriva based catalogs.  The module also implements the Feature and Vocabulary functions
in the DerivaML.

DerivaML and its associated classes all depend on a catalog that implements a `deriva-ml` schema with tables and
relationships that follow a specific data model.

"""

from deriva.core.ermrest_model import Table, Model
from deriva.core.ermrest_catalog import ErmrestCatalog

from .deriva_definitions import (
    DerivaMLException,
    ML_SCHEMA,
    DerivaSystemColumns,
    TableDefinition,
)

from pydantic import validate_call, ConfigDict


class DerivaModel:
    """Augmented interface to deriva model class.

    Attributes:
        domain_schema: Schema name for domain specific tables and relationships.
        model: ERMRest model for the catalog.
        schemas: ERMRest model for the catalog.
        catalog: ERMRest catalog for the model
        hostname: ERMRest catalog for the model

    """

    def __init__(
        self, model: Model, ml_schema: str = ML_SCHEMA, domain_schema: str = ""
    ):
        """Create and initialize a DerivaML instance.

        This method will connect to a catalog, and initialize local configuration for the ML execution.
        This class is intended to be used as a base class on which domain-specific interfaces are built.

        Args:
        """
        self.model = model
        self.configuration = None
        self.catalog: ErmrestCatalog = self.model.catalog
        self.hostname = (
            self.catalog.deriva_server.server
            if isinstance(self.catalog, ErmrestCatalog)
            else "localhost"
        )
        self.schemas = self.model.schemas

        self.ml_schema = ml_schema
        builtin_schemas = ["public", self.ml_schema, "www"]
        try:
            self.domain_schema = (
                domain_schema
                or [
                    s for s in self.model.schemas.keys() if s not in builtin_schemas
                ].pop()
            )
        except IndexError:
            # No domain schema defined.
            self.domain_schema = domain_schema

    def get_table(self, table: str | Table) -> Table:
        """Return the table object corresponding to the given table name.

        If the table name appears in more than one schema, return the first one you find.

        Args:
          table: A ERMRest table object or a string that is the name of the table.
          table: str | Table:

        Returns:
          Table object.
        """
        if isinstance(table, Table):
            return table
        for s in self.model.schemas.values():
            if table in s.tables.keys():
                return s.tables[table]
        raise DerivaMLException(f"The table {table} doesn't exist.")

    def is_vocabulary(self, table_name: str | Table) -> bool:
        """Check if a given table is a controlled vocabulary table.

        Args:
          table_name: A ERMRest table object or the name of the table.
          table_name: str | Table:

        Returns:
          Table object if the table is a controlled vocabulary, False otherwise.

        Raises:
          DerivaMLException: if the table doesn't exist.

        """
        vocab_columns = {"NAME", "URI", "SYNONYMS", "DESCRIPTION", "ID"}
        table = self.get_table(table_name)
        return vocab_columns.issubset({c.name.upper() for c in table.columns})

    def is_association(
        self, table_name: str | Table, unqualified: bool = True, pure: bool = True
    ) -> bool | set | int:
        """Check the specified table to see if it is an association table.

        Args:
            table_name: param unqualified:
            pure: return: (Default value = True)
            table_name: str | Table:
            unqualified:  (Default value = True)

        Returns:


        """
        table = self.get_table(table_name)
        return table.is_association(unqualified=unqualified, pure=pure)

    def is_asset(self, table_name: str | Table) -> bool:
        """True if the specified table is an asset table.

        Args:
            table_name: str | Table:

        Returns:
            True if the specified table is an asset table, False otherwise.

        """
        asset_columns = {"Filename", "URL", "Length", "MD5", "Description"}
        table = self.get_table(table_name)
        return asset_columns.issubset({c.name for c in table.columns})

    def find_assets(self, with_metadata: bool = False) -> list[Table]:
        """Return the list of asset tables in the current model"""
        return [
            t
            for s in self.model.schemas.values()
            for t in s.tables.values()
            if self.is_asset(t)
        ]

    def find_vocabularies(self) -> list[Table]:
        """Return a list of all the controlled vocabulary tables in the domain schema."""
        return [
            t
            for s in self.model.schemas.values()
            for t in s.tables.values()
            if self.is_vocabulary(t)
        ]

    def asset_metadata(self, table: str | Table) -> set[str]:
        """Return the metadata columns for an asset table."""

        table = self.get_table(table)
        asset_columns = {
            "Filename",
            "URL",
            "Length",
            "MD5",
            "Description",
        }.union(set(DerivaSystemColumns))

        if not self.is_asset(table):
            raise DerivaMLException(f"{table.name} is not an asset table.")
        return {c.name for c in table.columns} - asset_columns

    def apply(self):
        """Call Ermrestmodel.apply"""
        if self.catalog == "file-system":
            raise DerivaMLException("Cannot apply() to non-catalog model.")
        else:
            self.model.apply()

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def create_table(self, table_def: TableDefinition) -> Table:
        """Create a new table from TableDefinition."""
        return self.model.schemas[self.domain_schema].create_table(
            table_def.model_dump()
        )



import argparse
import sys

from deriva.core import DerivaServer, ErmrestCatalog, get_credential
from deriva.core.ermrest_model import Model
from deriva.core.ermrest_model import (
    builtin_types,
    Schema,
    Table,
    Column,
    ForeignKey,
    Key,
)
from deriva.core.utils.core_utils import tag as chaise_tags

from deriva_ml import MLVocab
from deriva_ml.schema_setup.annotations import generate_annotation
from deriva_ml.deriva_model import DerivaModel


def define_table_workflow(workflow_annotation: dict):
    return Table.define(
        "Workflow",
        column_defs=[
            Column.define("Name", builtin_types.text),
            Column.define("Description", builtin_types.markdown),
            Column.define("URL", builtin_types.ermrest_uri),
            Column.define("Checksum", builtin_types.text),
            Column.define("Version", builtin_types.text),
        ],
        annotations=workflow_annotation,
    )


def define_table_dataset(sname, dataset_annotation: dict = None):
    return Table.define(
        tname="Dataset",
        column_defs=[
            Column.define("Description", builtin_types.markdown),
            Column.define("Deleted", builtin_types.boolean),
        ],
        annotations=dataset_annotation if dataset_annotation is not None else {},
    )


def define_table_dataset_version(sname: str, dataset_version_annotation: dict = None):
    return Table.define(
        tname="Dataset_Version",
        column_defs=[
            Column.define(
                "Version",
                builtin_types.text,
                default="0.1.0",
                comment="Semantic version of dataset",
            ),
            Column.define("Description", builtin_types.markdown),
            Column.define("Dataset", builtin_types.text, comment="RID of dataset"),
            Column.define(
                "Minid", builtin_types.text, comment="URL to MINID for dataset"
            ),
        ],
        key_defs=[Key.define(["Dataset", "Version"])],
        fkey_defs=[ForeignKey.define(["Dataset"], sname, "Dataset", ["RID"])],
    )


def define_table_execution(sname: str, execution_annotation: dict):
    table_def = Table.define(
        "Execution",
        column_defs=[
            Column.define("Workflow", builtin_types.text),
            Column.define("Description", builtin_types.markdown),
            Column.define("Duration", builtin_types.text),
            Column.define("Status", builtin_types.text),
            Column.define("Status_Detail", builtin_types.text),
        ],
        fkey_defs=[ForeignKey.define(["Workflow"], sname, "Workflow", ["RID"])],
        annotations=execution_annotation,
    )
    return table_def


def define_asset_execution_metadata(sname: str, execution_metadata_annotation: dict):
    return Table.define_asset(
        sname=sname,
        tname="Execution_Metadata",
        column_defs=[Column.define("Execution", builtin_types.markdown)],
        hatrac_template="/hatrac/metadata/{{MD5}}.{{Filename}}",
        fkey_defs=[ForeignKey.define(["Execution"], sname, "Execution", ["RID"])],
        annotations=execution_metadata_annotation,
    )


def define_asset_execution_asset(sname: str, execution_asset_annotation: dict):
    table_def = Table.define_asset(
        sname=sname,
        tname="Execution_Asset",
        hatrac_template="/hatrac/execution_asset/{{MD5}}.{{Filename}}",
        annotations=execution_asset_annotation,
    )
    return table_def


def create_www_schema(model: Model):
    """
    Set up a new schema and tables to hold web-page like content.  The tables include a page table, and an asset
    table that can have images that are referred to by the web page.  Pages are written using markdown.
    :return:
    """
    if model.schemas.get("www"):
        model.schemas["www"].drop(cascade=True)
    www_schema = model.create_schema(
        Schema.define(
            "www", comment="Schema for tables that will be displayed as web content"
        )
    )
    www_schema.create_table(
        Table.define(
            "Page",
            column_defs=[
                Column.define(
                    "Title",
                    builtin_types.text,
                    nullok=False,
                    comment="Unique title for the page",
                ),
                Column.define(
                    "Content",
                    builtin_types.markdown,
                    comment="Content of the page in markdown",
                ),
            ],
            key_defs=[Key.define(["Title"])],
            annotations={
                chaise_tags.table_display: {
                    "detailed": {
                        "hide_column_headers": True,
                        "collapse_toc_panel": True,
                    }
                },
                chaise_tags.visible_foreign_keys: {"detailed": {}},
                chaise_tags.visible_columns: {"detailed": ["Content"]},
            },
        )
    )

    return www_schema


def create_ml_schema(
    model: Model, schema_name: str = "deriva-ml", project_name: str = None
):
    ml_catalog: ErmrestCatalog = model.catalog

    if model.schemas.get(schema_name):
        model.schemas[schema_name].drop(cascade=True)
    # get annotations
    deriva_model = DerivaModel(model)
    annotations = generate_annotation(deriva_model)

    model.annotations.update(annotations["catalog_annotation"])
    client_annotation = {
        "tag:misd.isi.edu,2015:display": {"name": "Users"},
        "tag:isrd.isi.edu,2016:table-display": {
            "row_name": {"row_markdown_pattern": "{{{Full_Name}}}"}
        },
        "tag:isrd.isi.edu,2016:visible-columns": {
            "compact": ["Full_Name", "Display_Name", "Email", "ID"]
        },
    }
    model.schemas["public"].tables["ERMrest_Client"].annotations.update(
        client_annotation
    )
    model.apply()

    schema = model.create_schema(
        Schema.define(schema_name, annotations=annotations["schema_annotation"])
    )
    project_name = project_name or schema_name
    # Workflow
    schema.create_table(
        Table.define_vocabulary("Feature_Name", f"{project_name}:{{RID}}")
    )

    workflow_table = schema.create_table(
        define_table_workflow(annotations["workflow_annotation"])
    )
    workflow_table.create_reference(
        schema.create_table(
            Table.define_vocabulary(MLVocab.workflow_type, f"{schema_name}:{{RID}}")
        )
    )

    execution_table = schema.create_table(
        define_table_execution(schema_name, annotations["execution_annotation"])
    )

    dataset_table = schema.create_table(
        define_table_dataset(schema_name, annotations["dataset_annotation"])
    )
    dataset_type = schema.create_table(
        Table.define_vocabulary(MLVocab.dataset_type, f"{project_name}:{{RID}}")
    )
    schema.create_table(
        Table.define_association(
            associates=[
                ("Dataset", dataset_table),
                (MLVocab.dataset_type, dataset_type),
            ]
        )
    )
    schema.create_table(
        Table.define_association(
            associates=[("Dataset", dataset_table), ("Execution", execution_table)]
        )
    )

    dataset_version = schema.create_table(define_table_dataset_version(schema_name))
    dataset_table.create_reference(("Version", True, dataset_version))

    # Nested datasets.
    schema.create_table(
        Table.define_association(
            associates=[("Dataset", dataset_table), ("Nested_Dataset", dataset_table)]
        )
    )

    # Execution Metadata
    execution_metadata_table = schema.create_table(
        define_asset_execution_metadata(
            schema.name, annotations["execution_metadata_annotation"]
        )
    )
    execution_metadata_table.create_reference(
        schema.create_table(
            Table.define_vocabulary(
                "Execution_Metadata_Type", f"{project_name}:{{RID}}"
            )
        )
    )
    schema.create_table(
        Table.define_association(
            [
                ("Execution_Metadata", execution_metadata_table),
                ("Execution", execution_table),
            ]
        )
    )

    # Execution Asset
    execution_asset_table = schema.create_table(
        define_asset_execution_asset(
            schema.name, annotations["execution_asset_annotation"]
        )
    )
    execution_asset_table.create_reference(
        schema.create_table(
            Table.define_vocabulary("Execution_Asset_Type", f"{project_name}:{{RID}}")
        )
    )
    schema.create_table(
        Table.define_association(
            [("Execution_Asset", execution_asset_table), ("Execution", execution_table)]
        )
    )

    create_www_schema(model)
    initialize_ml_schema(model, schema_name)


def initialize_ml_schema(model: Model, schema_name: str = "deriva-ml"):
    catalog = model.catalog
    execution_metadata_type = (
        catalog.getPathBuilder().schemas[schema_name].tables["Execution_Metadata_Type"]
    )
    execution_metadata_type.insert(
        [
            {
                "Name": "Execution_Config",
                "Description": "Configuration File for execution metadata",
            },
            {
                "Name": "Runtime_Env",
                "Description": "Information about the execution environment",
            },
        ],
        defaults={"ID", "URI"},
    )


def main():
    scheme = "https"
    parser = argparse.ArgumentParser()
    parser.add_argument("--hostname", type=str, required=True)
    parser.add_argument("--schema_name", type=str, required=True)
    parser.add_argument("--catalog_id", type=str, required=True)
    parser.add_argument("--curie_prefix", type=str, required=True)
    args = parser.parse_args()
    credentials = get_credential(args.hostname)
    server = DerivaServer(scheme, args.hostname, credentials)
    model = server.connect_ermrest(args.catalog_id).getCatalogModel()
    create_ml_schema(model, args.schema_name)


if __name__ == "__main__":
    sys.exit(main())



import logging
import sqlite3

from csv import reader
from pathlib import Path
from typing import Any, Generator, Optional
from urllib.parse import urlparse

import pandas as pd
from deriva.core.ermrest_model import Model
from pydantic import validate_call

from .deriva_definitions import ML_SCHEMA, MLVocab, RID, DerivaMLException
from .dataset_aux_classes import DatasetVersion, DatasetMinid
from .deriva_model import DerivaModel
from .dataset_bag import DatasetBag


class DatabaseModel(DerivaModel):
    """Read in the contents of a BDBag and create a local SQLite database.

        As part of its initialization, this routine will create a sqlite database that has the contents of all the tables
    in the dataset_table.  In addition, any asset tables will the `Filename` column remapped to have the path of the local
    copy of the file. In addition, a local version of the ERMRest model that as used to generate the dataset_table is
    available.

       The sqlite database will not have any foreign key constraints applied, however, foreign-key relationships can be
    found by looking in the ERMRest model.  In addition, as sqllite doesn't support schema, Ermrest schema are added
    to the table name using the convention SchemaName:TableName.  Methods in DatasetBag that have table names as the
    argument will perform the appropriate name mappings.

    Because of nested datasets, it's possible that more than one dataset rid is in a bag, or that a dataset rid might
    appear in more than one database. To help manage this, a global list of all the datasets that have been loaded
    into DatabaseModels, is kept in the class variable `_rid_map`.

    Attributes:
        bag_path (Path): path to the local copy of the BDBag
        minid (DatasetMinid): Minid for the specified bag
        dataset_rid (RID): RID for the specified dataset
        dbase (Connection): connection to the sqlite database holding table values
        domain_schema (str): Name of the domain schema
        dataset_table  (Table): the dataset table in the ERMRest model.
    """

    # Keep track of what databases we have loaded.
    _paths_loaded: dict[Path:"DatabaseModel"] = {}

    # Maintain a global map of RIDS to versions and databases.
    _rid_map: dict[RID, list[tuple[DatasetVersion, "DatabaseModel"]]] = {}

    @classmethod
    @validate_call
    def register(cls, minid: DatasetMinid, bag_path: Path):
        """Register a new minid in the list of local databases if it's new, otherwise, return an existing DatabaseModel.

        Args:
            minid: MINID to the databag that is to be loaded.
            bag_path: Path to the bag on the local filesystem./

        Returns:
            A DatabaseModel instance to the loaded bag.
        """
        o = cls._paths_loaded.get(bag_path.as_posix())
        if o:
            return o
        return cls(minid, bag_path)

    @staticmethod
    def rid_lookup(dataset_rid: RID) -> list[tuple[DatasetVersion, "DatabaseModel"]]:
        """Return a list of DatasetVersion/DatabaseModel instances corresponding to the given RID.

        Args:
            dataset_rid: Rit to be looked up.

        Returns:
            List of DatasetVersion/DatabaseModel instances corresponding to the given RID.

        Raises:
            Raise a DerivaMLException if the given RID is not found.
        """
        try:
            return DatabaseModel._rid_map[dataset_rid]
        except KeyError:
            raise DerivaMLException(f"Dataset {dataset_rid} not found")

    def __init__(self, minid: DatasetMinid, bag_path: Path):
        """Create a new DatabaseModel.  This should only be called via the static Register method

        Args:
            minid: Minid for the specified bag.
            bag_path:  Path to the local copy of the BDBag.
        """
        DatabaseModel._paths_loaded[bag_path.as_posix()] = self

        self.bag_path = bag_path
        self.minid = minid
        self.dataset_rid = minid.dataset_rid
        dir_path = bag_path.parent
        self.dbase_file = dir_path / f"{minid.version_rid}.db"
        self.dbase = sqlite3.connect(self.dbase_file)

        super().__init__(
            Model.fromfile("file-system", self.bag_path / "data/schema.json")
        )

        self._logger = logging.getLogger("deriva_ml")
        self.domain_schema = self._guess_domain_schema()
        self._load_model()
        self.ml_schema = ML_SCHEMA
        self._load_sqlite()
        self._logger.info(
            "Creating new database for dataset: %s in %s",
            self.dataset_rid,
            self.dbase_file,
        )
        self.dataset_table = self.model.schemas[self.ml_schema].tables["Dataset"]
        # Now go through the database and pick out all the dataset_table RIDS, along with their versions.
        sql_dataset = self.normalize_table_name("Dataset_Version")
        with self.dbase:
            dataset_versions = [
                t
                for t in self.dbase.execute(
                    f'SELECT "Dataset", "Version" FROM "{sql_dataset}"'
                ).fetchall()
            ]
        dataset_versions = [
            (v[0], DatasetVersion.parse(v[1])) for v in dataset_versions
        ]

        # Get most current version of each rid
        self.bag_rids = {}
        for rid, version in dataset_versions:
            self.bag_rids[rid] = max(
                self.bag_rids.get(rid, DatasetVersion(0, 1, 0)), version
            )

        for dataset_rid, dataset_version in self.bag_rids.items():
            version_list = DatabaseModel._rid_map.setdefault(dataset_rid, [])
            version_list.append((dataset_version, self))

    def _load_model(self) -> None:
        """Create a sqlite database schema that contains all the tables within the catalog from which the BDBag was created."""
        with self.dbase:
            for t in self.model.schemas[self.domain_schema].tables.values():
                self.dbase.execute(t.sqlite3_ddl())
            for t in self.model.schemas["deriva-ml"].tables.values():
                self.dbase.execute(t.sqlite3_ddl())

    def _load_sqlite(self) -> None:
        """Load a SQLite database from a bdbag.  THis is done by looking for all the CSV files in the bdbag directory.

        If the file is for an asset table, update the FileName column of the table to have the local file path for
        the materialized file.  Then load into the sqllite database.
        Note: none of the foreign key constraints are included in the database.
        """
        dpath = self.bag_path / "data"
        asset_map = self._localize_asset_table()  # Map of remote to local assets.

        # Find all the CSV files in the subdirectory and load each file into the database.
        for csv_file in Path(dpath).rglob("*.csv"):
            table = csv_file.stem
            schema = (
                self.domain_schema
                if table in self.model.schemas[self.domain_schema].tables
                else self.ml_schema
            )

            with csv_file.open(newline="") as csvfile:
                csv_reader = reader(csvfile)
                column_names = next(csv_reader)

                # Determine which columns in the table has the Filename and the URL
                asset_indexes = (
                    (column_names.index("Filename"), column_names.index("URL"))
                    if self._is_asset(table)
                    else None
                )

                value_template = ",".join(
                    ["?"] * len(column_names)
                )  # SQL placeholder for row (?,?..)
                column_list = ",".join([f'"{c}"' for c in column_names])
                with self.dbase:
                    object_table = (
                        self._localize_asset(o, asset_indexes, asset_map)
                        for o in csv_reader
                    )
                    self.dbase.executemany(
                        f'INSERT OR REPLACE INTO "{schema}:{table}" ({column_list}) VALUES ({value_template})',
                        object_table,
                    )

    def _localize_asset_table(self) -> dict[str, str]:
        """Use the fetch.txt file in a bdbag to create a map from a URL to a local file path.

        Returns:
            Dictionary that maps a URL to a local file path.

        """
        fetch_map = {}
        try:
            with open(self.bag_path / "fetch.txt", newline="\n") as fetch_file:
                for row in fetch_file:
                    # Rows in fetch.text are tab seperated with URL filename.
                    fields = row.split("\t")
                    local_file = fields[2].replace("\n", "")
                    local_path = f"{self.bag_path}/{local_file}"
                    fetch_map[urlparse(fields[0]).path] = local_path
        except FileNotFoundError:
            dataset_rid = self.bag_path.name.replace("Dataset_", "")
            logging.info(f"No downloaded assets in bag {dataset_rid}")
        return fetch_map

    def _guess_domain_schema(self) -> str:
        """Guess the domain schema name by eliminating all the "builtin" schema.

        Returns:
            String for domain schema name.
        """
        return [
            s for s in self.model.schemas if s not in ["deriva-ml", "public", "www"]
        ][0]

    def _is_asset(self, table_name: str) -> bool:
        """

        Args:
          table_name: str:

        Returns:
            Boolean that is true if the table looks like an asset table.
        """
        asset_columns = {"Filename", "URL", "Length", "MD5", "Description"}
        sname = (
            self.domain_schema
            if table_name in self.model.schemas[self.domain_schema].tables
            else self.ml_schema
        )
        asset_table = self.model.schemas[sname].tables[table_name]
        return asset_columns.issubset({c.name for c in asset_table.columns})

    @staticmethod
    def _localize_asset(
        o: list, indexes: tuple[int, int], asset_map: dict[str, str]
    ) -> tuple:
        """Given a list of column values for a table, replace the FileName column with the local file name based on
        the URL value.

        Args:
          o: List of values for each column in a table row.
          indexes: A tuple whose first element is the column index of the file name and whose second element
        is the index of the URL in an asset table.  Tuple is None if table is not an asset table.
          o: list:
          indexes: Optional[tuple[int: int]]:

        Returns:
          Tuple of updated column values.

        """
        if indexes:
            file_column, url_column = indexes
            o[file_column] = asset_map[o[url_column]] if o[url_column] else ""
        return tuple(o)

    def list_tables(self) -> list[str]:
        """List the names of the tables in the catalog

        Returns:
            A list of table names.  These names are all qualified with the Deriva schema name.
        """
        with self.dbase:
            return [
                t[0]
                for t in self.dbase.execute(
                    "SELECT name FROM sqlite_master WHERE type = 'table' ORDER BY name;"
                ).fetchall()
            ]

    def get_dataset(self, dataset_rid: Optional[RID] = None) -> DatasetBag:
        """Get a dataset, or nested dataset from the bag database

        Args:
            dataset_rid: Optional.  If not provided, use the main RID for the bag.  If a value is given, it must
            be the RID for a nested dataset.

        Returns:
            DatasetBag object for the specified dataset.
        """
        if dataset_rid and dataset_rid not in self.bag_rids:
            DerivaMLException(f"Dataset RID {dataset_rid} is not in model.")
        return DatasetBag(self, dataset_rid or self.dataset_rid)

    def dataset_version(self, dataset_rid: Optional[RID] = None) -> DatasetVersion:
        if dataset_rid and dataset_rid not in self.bag_rids:
            DerivaMLException(f"Dataset RID {dataset_rid} is not in model.")
        return self.bag_rids[dataset_rid]

    def find_datasets(self) -> list[dict[str, Any]]:
        """Returns a list of currently available datasets.

        Returns:
             list of currently available datasets.
        """
        atable = next(
            self.model.schemas[ML_SCHEMA]
            .tables[MLVocab.dataset_type]
            .find_associations()
        ).name

        # Get a list of all the dataset_type values associated with this dataset_table.
        datasets = []
        ds_types = list(self.get_table_as_dict(atable))
        for dataset in self.get_table_as_dict("Dataset"):
            my_types = [t for t in ds_types if t["Dataset"] == dataset["RID"]]
            datasets.append(
                dataset
                | {MLVocab.dataset_type: [ds[MLVocab.dataset_type] for ds in my_types]}
            )
        return datasets

    def normalize_table_name(self, table: str) -> str:
        """Attempt to insert the schema into a table name if it's not provided.

        Args:
          table: str:

        Returns:
          table name with schema included.

        """
        sname = ""
        try:
            [sname, tname] = table.split(":")
        except ValueError:
            tname = table
            for sname, s in self.model.schemas.items():
                if table in s.tables:
                    break
        try:
            _ = self.model.schemas[sname].tables[tname]
            return f"{sname}:{tname}"
        except KeyError:
            raise DerivaMLException(f'Table name "{table}" does not exist.')

    def get_table(self, table: str) -> Generator[tuple, None, None]:
        """Retrieve the contents of the specified table. If schema is not provided as part of the table name,
        the method will attempt to locate the schema for the table.

        Args:
            table: return: A generator that yields tuples of column values.

        Returns:
          A generator that yields tuples of column values.

        """
        table_name = self.normalize_table_name(table)
        result = self.dbase.execute(f'SELECT * FROM "{table_name}"')
        while row := result.fetchone():
            yield row

    def get_table_as_dataframe(self, table: str) -> pd.DataFrame:
        """Retrieve the contents of the specified table as a dataframe.


        If schema is not provided as part of the table name,
        the method will attempt to locate the schema for the table.

        Args:
            table: Table to retrieve data from.

        Returns:
          A dataframe containing the contents of the specified table.
        """
        table_name = self.normalize_table_name(table)
        return pd.read_sql(f'SELECT * FROM "{table_name}"', con=self.dbase)

    def get_table_as_dict(self, table: str) -> Generator[dict[str, Any], None, None]:
        """Retrieve the contents of the specified table as a dictionary.

        Args:
            table: Table to retrieve data from. f schema is not provided as part of the table name,
                the method will attempt to locate the schema for the table.

        Returns:
          A generator producing dictionaries containing the contents of the specified table as name/value pairs.
        """
        table_name = self.normalize_table_name(table)
        with self.dbase:
            col_names = [
                c[1]
                for c in self.dbase.execute(
                    f'PRAGMA table_info("{table_name}")'
                ).fetchall()
            ]
            result = self.dbase.execute(f'SELECT * FROM "{table_name}"')
            while row := result.fetchone():
                yield dict(zip(col_names, row))

    def delete_database(self):
        """

        Args:

        Returns:

        """
        self.dbase_file.unlink()



import atexit
from importlib.metadata import version
from importlib.resources import files
import logging
from random import random
from tempfile import TemporaryDirectory

from deriva.config.acl_config import AclConfig
from deriva.core import DerivaServer
from deriva.core import ErmrestCatalog, get_credential
from deriva.core.datapath import DataPathException
from deriva.core.ermrest_model import Model
from deriva.core.ermrest_model import builtin_types, Schema, Table, Column
from requests import HTTPError

from deriva_ml import (
    DerivaML,
    ExecutionConfiguration,
    Workflow,
    MLVocab,
    BuiltinTypes,
    ColumnDefinition,
)

from deriva_ml.execution import Execution
from deriva_ml.schema_setup.create_schema import initialize_ml_schema, create_ml_schema
from deriva_ml.dataset import Dataset

TEST_DATASET_SIZE = 4


def reset_demo_catalog(deriva_ml: DerivaML, sname: str):
    model = deriva_ml.model
    for trial in range(3):
        for t in [
            v
            for v in model.schemas[sname].tables.values()
            if v.name not in {"Subject", "Image"}
        ]:
            try:
                t.drop()
            except HTTPError:
                pass

    # Empty out remaining tables.
    pb = deriva_ml.pathBuilder
    retry = True
    while retry:
        retry = False
        for s in [sname, "deriva-ml"]:
            for t in pb.schemas[s].tables.values():
                for e in t.entities().fetch():
                    try:
                        t.filter(t.RID == e["RID"]).delete()
                    except DataPathException:  # FK constraint.
                        retry = True

    initialize_ml_schema(model, "deriva-ml")


def populate_demo_catalog(deriva_ml: DerivaML, sname: str) -> None:
    # Delete any vocabularies and features.
    reset_demo_catalog(deriva_ml, sname)
    domain_schema = deriva_ml.catalog.getPathBuilder().schemas[sname]
    subject = domain_schema.tables["Subject"]
    ss = subject.insert([{"Name": f"Thing{t + 1}"} for t in range(TEST_DATASET_SIZE)])

    with TemporaryDirectory() as tmpdir:
        image_dir = deriva_ml.asset_dir("Image", prefix=tmpdir)
        for s in ss:
            image_file = image_dir.create_file(
                f"test_{s['RID']}.txt", {"Subject": s["RID"]}
            )
            with open(image_file, "w") as f:
                f.write(f"Hello there {random()}\n")
        deriva_ml.upload_assets(image_dir)


def create_demo_datasets(deriva_ml: DerivaML) -> None:
    deriva_ml.add_dataset_element_type("Subject")
    deriva_ml.add_dataset_element_type("Image")

    # Create a new dataset_table
    deriva_ml.add_term(
        MLVocab.dataset_type, "DemoSet", description="A test dataset_table"
    )
    deriva_ml.add_term(
        MLVocab.dataset_type,
        "Partitioned",
        description="A partitioned dataset_table for ML training.",
    )
    deriva_ml.add_term(
        MLVocab.dataset_type, "Subject", description="A test dataset_table"
    )
    deriva_ml.add_term(
        MLVocab.dataset_type, "Image", description="A test dataset_table"
    )
    deriva_ml.add_term(
        MLVocab.dataset_type, "Training", description="Training dataset_table"
    )
    deriva_ml.add_term(
        MLVocab.dataset_type, "Testing", description="Training dataset_table"
    )
    deriva_ml.add_term(
        MLVocab.dataset_type, "Validation", description="Validation dataset_table"
    )

    deriva_ml.add_term(
        MLVocab.workflow_type,
        "Create Dataset Notebook",
        description="A Workflow that creates a new dataset_table",
    )

    # Now let's create model configuration for our program.
    api_workflow = Workflow(
        name="API Workflow",
        url="https://github.com/informatics-isi-edu/deriva-ml/blob/main/src/deriva_ml/demo_catalog.py",
        workflow_type="Create Dataset Notebook",
    )

    dataset_execution = Execution(
        ExecutionConfiguration(
            workflow=api_workflow, description="Create demo dataset_table"
        ),
        deriva_ml,
    )

    pb = deriva_ml.pathBuilder.schemas[deriva_ml.domain_schema]
    subject_path = pb.Subject
    image_path = pb.Image

    # Get images and subjects.  Join images with subjects.
    images = {
        i["Subject"]: i["RID"]
        for i in image_path.attributes(image_path.RID, image_path.Subject).fetch()
    }
    subjects = [
        s | {"Image": images[s["RID"]]}
        for s in subject_path.attributes(subject_path.RID, subject_path.Name).fetch()
    ]

    def thing_number(name: str) -> int:
        return int(name.replace("Thing", ""))

    training_rids = [s["Image"] for s in subjects if thing_number(s["Name"]) % 3 == 0]
    testing_rids = [s["Image"] for s in subjects if thing_number(s["Name"]) % 3 == 1]
    validation_rids = [s["Image"] for s in subjects if thing_number(s["Name"]) % 3 == 2]

    nested_dataset = deriva_ml.create_dataset(
        ["Partitioned", "Image"],
        description="A nested dataset_table for machine learning",
        execution_rid=dataset_execution.execution_rid,
    )
    double_nested_dataset = deriva_ml.create_dataset(
        ["Partitioned", "Image"],
        description="A double nested dataset_table for machine learning",
        execution_rid=dataset_execution.execution_rid,
    )
    training_dataset = deriva_ml.create_dataset(
        "Training",
        description="An image dataset_table for training",
        execution_rid=dataset_execution.execution_rid,
    )
    testing_dataset = deriva_ml.create_dataset(
        "Testing",
        description="A image dataset_table for testing",
        execution_rid=dataset_execution.execution_rid,
    )
    validation_dataset = deriva_ml.create_dataset(
        "Validation",
        description="A image dataset_table for validation",
        execution_rid=dataset_execution.execution_rid,
    )
    deriva_ml.add_dataset_members(
        dataset_rid=nested_dataset,
        members=[training_dataset, testing_dataset],
    )

    deriva_ml.add_dataset_members(
        dataset_rid=double_nested_dataset,
        members=[nested_dataset, validation_dataset],
    )
    deriva_ml.add_dataset_members(dataset_rid=training_dataset, members=training_rids)
    deriva_ml.add_dataset_members(dataset_rid=testing_dataset, members=testing_rids)
    deriva_ml.add_dataset_members(
        dataset_rid=validation_dataset, members=validation_rids
    )


def create_demo_features(deriva_ml: DerivaML) -> None:
    deriva_ml.create_vocabulary("SubjectHealth", "A vocab")
    deriva_ml.add_term(
        "SubjectHealth",
        "Sick",
        description="The subject self reports that they are sick",
    )
    deriva_ml.add_term(
        "SubjectHealth",
        "Well",
        description="The subject self reports that they feel well",
    )

    deriva_ml.create_vocabulary(
        "ImageQuality", "Controlled vocabulary for image quality"
    )
    deriva_ml.add_term("ImageQuality", "Good", description="The image is good")
    deriva_ml.add_term("ImageQuality", "Bad", description="The image is bad")

    box_asset = deriva_ml.create_asset(
        "BoundingBox", comment="A file that contains a cropped version of a image"
    )

    deriva_ml.create_feature(
        "Subject",
        "Health",
        terms=["SubjectHealth"],
        metadata=[ColumnDefinition(name="Scale", type=BuiltinTypes.int2, nullok=True)],
        optional=["Scale"],
    )

    deriva_ml.create_feature("Image", "BoundingBox", assets=[box_asset])
    deriva_ml.create_feature("Image", "Quality", terms=["ImageQuality"])


def create_domain_schema(model: Model, sname: str) -> None:
    """
    Create a domain schema.  Assumes that the ml-schema has already been created.
    :param model:
    :param sname:
    :return:
    """

    # Make sure that we have a ml schema
    _ = model.schemas["deriva-ml"]

    if model.schemas.get(sname):
        # Clean out any old junk....
        model.schemas[sname].drop()

    domain_schema = model.create_schema(
        Schema.define(sname, annotations={"name_style": {"underline_space": True}})
    )
    subject_table = domain_schema.create_table(
        Table.define("Subject", column_defs=[Column.define("Name", builtin_types.text)])
    )

    image_table = domain_schema.create_table(
        Table.define_asset(
            sname=sname,
            tname="Image",
            hatrac_template="/hatrac/image_asset/{{MD5}}.{{Filename}}",
            column_defs=[Column.define("Name", builtin_types.text)],
        )
    )
    image_table.create_reference(subject_table)


def destroy_demo_catalog(catalog):
    catalog.delete_ermrest_catalog(really=True)


def create_demo_catalog(
    hostname,
    domain_schema="test-schema",
    project_name="ml-test",
    populate=True,
    create_features=False,
    create_datasets=False,
    on_exit_delete=True,
) -> ErmrestCatalog:
    credentials = get_credential(hostname)
    server = DerivaServer("https", hostname, credentials=credentials)
    test_catalog = server.create_ermrest_catalog()
    if on_exit_delete:
        atexit.register(destroy_demo_catalog, test_catalog)
    model = test_catalog.getCatalogModel()

    try:
        create_ml_schema(model, project_name=project_name)
        create_domain_schema(model, domain_schema)
        deriva_ml = DerivaML(
            hostname=hostname,
            catalog_id=test_catalog.catalog_id,
            project_name=project_name,
            logging_level=logging.WARN,
        )
        if populate:
            populate_demo_catalog(deriva_ml, domain_schema)
        if create_features:
            create_demo_features(deriva_ml)
        if create_datasets:
            create_demo_datasets(deriva_ml)
        dataset_table = deriva_ml.dataset_table
        dataset_table.annotations.update(
            Dataset(
                deriva_ml.model, deriva_ml.cache_dir
            )._generate_dataset_annotations()
        )
        deriva_ml.model.apply()
        policy_file = files("deriva_ml.schema_setup").joinpath("policy.json")
        AclConfig(
            hostname, test_catalog.catalog_id, policy_file, credentials=credentials
        )

    except Exception:
        # on failure, delete catalog and re-raise exception
        test_catalog.delete_ermrest_catalog(really=True)
        raise
    return test_catalog


class DemoML(DerivaML):
    def __init__(
        self, hostname, catalog_id, cache_dir: str = None, working_dir: str = None
    ):
        super().__init__(
            hostname=hostname,
            catalog_id=catalog_id,
            project_name="ml-test",
            cache_dir=cache_dir,
            working_dir=working_dir,
            model_version=version(__name__.split(".")[0]),
        )



from derivaml_test import TestDerivaML
from deriva_ml.demo_catalog import (
    reset_demo_catalog,
    populate_demo_catalog,
    create_demo_datasets,
)
from deriva_ml import DatasetSpec
from pathlib import Path


class TestFeatures(TestDerivaML):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        self.ml_instance = DerivaML(
            hostname, test_catalog.catalog_id, SNAME_DOMAIN, "", "", "1"
        )
        self.domain_schema = self.ml_instance.model.schemas[SNAME_DOMAIN]
        self.model = self.ml_instance.model

    def test_create_feature(self):
        populate_test_catalog(self.ml_instance, SNAME_DOMAIN)
        self.ml_instance.create_vocabulary("FeatureValue", "A vocab")
        self.ml_instance.add_term("FeatureValue", "V1", description="A Feature Vale")

        a = self.ml_instance.create_asset("TestAsset", comment="A asset")

        self.ml_instance.create_feature(
            "Feature1",
            "Image",
            terms=["FeatureValue"],
            assets=[a],
            metadata=[ColumnDefinition(name="TestCol", type=BuiltinTypes.int2)],
        )
        self.assertIn(
            "Execution_Image_Feature1",
            [f.name for f in self.ml_instance.find_features("Image")],
        )

    def test_add_feature(self):
        self.test_create_feature()
        TestFeature = self.ml_instance.feature_record_class("Image", "Feature1")
        # Create the name for this feature and then create the feature.
        # Get some images to attach the feature value to.
        domain_path = self.ml_instance.catalog.getPathBuilder().schemas[SNAME_DOMAIN]
        image_rids = [i["RID"] for i in domain_path.tables["Image"].entities().fetch()]
        asset_rid = domain_path.tables["TestAsset"].insert(
            [{"Name": "foo", "URL": "foo/bar", "Length": 2, "MD5": 4}]
        )[0]["RID"]
        # Get an execution RID.
        ml_path = self.ml_instance.catalog.getPathBuilder().schemas["deriva-ml"]
        self.ml_instance.add_term(
            "Workflow_Type", "TestWorkflow", description="A workflow"
        )
        workflow_rid = ml_path.tables["Workflow"].insert(
            [{"Name": "Test Workflow", "Workflow_Type": "TestWorkflow"}]
        )[0]["RID"]
        execution_rid = ml_path.tables["Execution"].insert(
            [{"Description": "Test execution", "Workflow": workflow_rid}]
        )[0]["RID"]
        # Now create a list of features using the feature creation class returned by create_feature.
        feature_list = [
            TestFeature(
                Image=i,
                Execution=execution_rid,
                FeatureValue="V1",
                TestAsset=asset_rid,
                TestCol=23,
            )
            for i in image_rids
        ]
        self.ml_instance.add_features(feature_list)
        features = self.ml_instance.list_feature_values("Image", "Feature1")
        self.assertEqual(len(features), len(image_rids))
